/*  Copyright (C) 2000, 2006  Marc Toussaint (mtoussai@inf.ed.ac.uk)
    under the terms of the GNU LGPL (http://www.gnu.org/copyleft/lesser.html)
    see the `std.h' file for a full copyright statement  */

/*  This implements the online LWPR regression algorithm as described in:
    Sethu Vijayakumar, Aaron D'Souza and Stefan Schaal, "Incremental Online
    Learning in High Dimensions", Neural Computation, 17:12, pp. 2602-2632  (2005)
    (Please cite when using code) */

#ifndef GP_LWPR_H
#define GP_LWPR_H

#include "array.h"


//forward declaration
class Lmodel;

/*!\brief computes the standard deviations along each dimension of an
    (input) data set and stores them in the \c norm vector --- which
    can be an Lwpr::norm used as normalization during learning and
    prediction */
void getStdDeviations(doubleA& norm,const doubleA& X);



/*! This is an interface to LWPR, Locally Weighted Projection
    Regression. */
class Lwpr {
public:
  //! constructor, also loads parameters from parameter file or command line
  Lwpr();

  //! updates the LWPR model given a new datapoint x->y, or a batch of data
  void learn(const doubleA& X,const doubleA& Y);

  //! predicts the output Y for a given input X -- also for batch data
  void predict(const doubleA& X,doubleA& Y);
  
  //! get the confidence bounds (standard deviation) at some input point
  double confidence(const doubleA &x){
    doubleA nx;
    normalizeInput(nx,x);
    return ::sqrt(s2pred(nx));
  }
  
  //! saves the LWPR model in the file given as parameter
  void save(char *);

  //! loads the LWPR model from the file given as parameter
  void load(char *,bool alsoParameters=true);

  //! preliminary -- output useful readable state information
  void report(std::ostream& os);
  
  //! get the number of rfs for the given output dimension
  int get_rfs_no(int out_dim);

  //! average number of projection dimensions used in the local models
  double get_proj_average(int out_dim);

  /*!\brief determines the verbosity during learning, set to 0 for silence */
  int verbosity; 

  /*!\brief number of receptive fields */
  uint rfsno;    

  /*!\brief The normalization by which each input column is divided
     (one entry for every input dimension) - this tries to make the
     all the inputs dimensionless. The norm field can be a single
     scalar (1d-array with N=1), a vector to normalizes each input
     dimension differently (1d-array), or a general matrix to
     transform the input arbitrarily (2d-array). (In general, it is
     dangerous to divide each dimension by its variance without
     considering the physical properties of the input values since
     some input dimensions may be actually moving very little relative
     to its range. Ideally, one should know the range of possible
     inputs in each dimension and try to normalize each input by
     that)*/
  doubleA norm;

  /*!\brief Allow(disallow) the addition of projection direction (to
      the initial number of projections specified by init_n_reg) if
      deemed necessary. This acts as a debugging variable.*/
  bool add_proj;   

  /*!\brief Allow(disallow) the gradient descent up date of the Distance Metric D*/
  bool updateD; 

  /*!\brief Allow(disallow) normalization */
  bool useNorm; 

  /*!\brief If none of the local models (receptive fields) elicits
      activation greater than cutoff for a query point, that query is
      labeled as not being well supported.  We should not trust
      predictions for this data point (We get separate test statistics
      which include/exclude these queries). LWPR returns ZERO for this
      model, there is no unlimited extrapolation. */

  double cut_off; 
  /*!\brief If a training data elicits responses greater than w_prune
      from 2 local models, then, the one with the greater error is
      pruned. */

  double w_prune;
  /*!\brief A new local model is generated if no existing model
      elicits response (activation) greater than w_gen for a training
      data. */ 

  double w_gen; 
  /*!\brief Initial values of distance metric */

  double initD; 
  /*!\brief Initial forgetting factor */

  double init_lambda;
  /*!\brief Annealing constant for the forgetting factor */ 

  double tau_lambda; 
  /*!\brief Final forgetting factor */

  double final_lambda; 

  /*!\brief The mean squared error of the current regression dimension
      is compared against the previous one. Only if the ratio of the
      nMSEr-1/nMSEr < add_threshold, a new regression direction is
      added. This criterion is used in conjunction with some other
      checks to ensure that the decision is based on enough data
      support. */
  double add_threshold; 

  /*!\brief Allow (disallow) second order learning. */
  bool meta; 

  /*!\brief To be used if second order learning is enabled */
  double meta_rate; 

  /*!\brief Number of projections to start regression with (Usually
      start with 2, one for regression and the other to compare
      whether an additional dimension should be used. */
  int initR; 

  /*!\brief Distance Metric learning rate initialization for gradient
     descent (currently implements only equal rates along all
     dimensions, i.e. initializes all the learning rates with the same
     parameters. (If you see lots of large update warnings or
     instability in convergence, you have too large a learning rate
     (alpha). If the Mean Squared Error is decreasing but the
     convergence is slow, you might try increasing the learning
     rate.) */
  double alpha; 

  /*!\brief Multiplication factor for the regularization penalty term
      in the optimization functional */
  double gamma; 

  /*!\brief 1: Allows weighted mixing of results from overlapping
      receptive fields.  0 : Winner-take-all prediction. */
  bool blend; 

private:
  void initialize(uint inDim,uint outDim);
  double s2pred(const doubleA& x);
  double predict(const MT::Array<Lmodel>& rfs, const doubleA& x);
  void updaterfs(MT::Array<Lmodel>& rfs, const doubleA& x, double y);
  void normalizeInput(doubleA& normIn,const doubleA& unNormIn);

  bool isInitialized;
  uint in_dim;
  uint out_dim;		

  MT::Array<MT::Array<Lmodel> > models;  //main data structure!
};

//================================================================================

#ifndef MT_doxy
/* This class defines the local models used by Lwpr - for experts
   only. Please refer to the Lwpr class first. */
class Lmodel {
public:
  Lwpr *lwpr;
  
  doubleA center; // center of the model
  doubleA mean_x; // x00
  double  mean_y; // beta00
  doubleA W;	  // running sum of weights
  double  MyMSE_R;// don't know
  doubleA lambda; // forgetting parameter
  doubleA M;	  // cholesky decomposition of dist
  doubleA n_data; // number of data seen

  
  MT::Array<doubleA> uprojections;   //u projections
  MT::Array<doubleA> pprojections;   //p projections
  MT::Array<doubleA> sXresYres;
  doubleA betas;   // betas
  doubleA MSE;     // squared errors

  doubleA azz;
  doubleA azres;
  MT::Array<doubleA> axz;
  
  //distance update statistics
  double sum_e2;
  doubleA  sum_ecv2;
  doubleA  aH;
  doubleA  aG;
  double aE;
  
  //penalty
  double gamma;
  //learning param
  doubleA  alpha;
  
  doubleA  b;
  doubleA  h;
  
  double apk;
  doubleA  dist;

public:
  int R;   // number of projections
  bool trustworthy;
  bool degenerated; // used for pruning of numerically degenerated local models
  
public:
  Lmodel(){};
  Lmodel(Lwpr* _lwpr,const doubleA& x, double y);
  Lmodel(Lwpr* _lwpr,const Lmodel& model, const doubleA& x, double y);
  double activation(const doubleA& x);
  void updateMean(const doubleA& x, double y);
  void updateStat(double w);
  void updateError(const doubleA& x, double y);
  void updateError_matlab(double, double); // emulates the error measure that's used in Matlab's LWPR implementation
  void update(const doubleA& x, double y);
  double predict(const doubleA& x ) const;
  void  update_dist(const doubleA& x, double y, double e_cv, double e);
  double gete(const doubleA& x, double y);
  double gete_cv(const doubleA& x, double y);
  bool check_add_projection(const doubleA& x, double y);
  void add_projection(const doubleA& x, double y);
  void printProj();
  friend std::ostream& operator<<(std::ostream& outs, const Lmodel& lm);

  void save(std::ostream&);
  void load(Lwpr* _lwpr, std::istream&);
  
  double s2pred(const doubleA& x, double w);
  void updateApk(const doubleA& x,double w);
  const doubleA& getCenter() const{ return center; }
  
private :
  void computeZs(doubleA& zs, const doubleA& x);
  void loadParameters(uint N);
  void initialise(int N);
  //void reinitialize();
  doubleA convert_vec(const doubleA& x);
  void getdJdM(doubleA& dJdM,const doubleA& D,const doubleA& M,
    const doubleA& x, const doubleA& center,
    const doubleA& z, const doubleA& azz, double w,
    double W, double e_cv, double e, double gamma,
    const doubleA& derivative_ok, double& sum_dJ1dw,
    doubleA& dwdM);
  
  double getsum_dJ1dw(double e_cv, double e, double w, double W,
    const doubleA& z_vec, const doubleA& azz_vec,
    const doubleA& derivative_ok);
  void update_dist_stat(double e_cv, double w, const doubleA& z_vec,
    const doubleA& azz_vec, double transient_multiplier,
    const doubleA& derivative_ok);
  void getdJ2dM(doubleA& dJ2dM,const doubleA& D, const doubleA& M, double gamma);
  void getdwdM(doubleA& dwdM,const doubleA& M, const doubleA& x, 
    const doubleA& center, double w);
  void getdDdMkl(doubleA& dDdMkl,const doubleA& M, int k, int l);
  
  //meta methods
  void meta_update(double e_cv, double W, double w, double e, const doubleA& aH,
    double aE, const doubleA& z_vec, 
    const doubleA& azz_vec, const doubleA& derivative_ok,
    const doubleA& x, const doubleA& center, double dJ1dw,
    const doubleA& dwdM, double transient_multiplier, const doubleA& dJdM);
  void getdJ2dJ2dMdM(doubleA& dJ2dJ2dMdM,const doubleA& D, const doubleA& M, 
    double gamma);
  void getdwdwdMdM(doubleA& dwdwdMdM,const doubleA& M, const doubleA& x, 
    const doubleA& center, double w);
  
  void cholesky(doubleA& L,const doubleA& A);
  intA find(const doubleA& A, double value, bool gt);
  double absmax(const doubleA& A);
};
#endif

#ifdef MT_IMPLEMENTATION
#include "lwpr.cpp"
#endif

#endif
